/*
 * ============================================================================
 *  Title:    Controller pads
 *  Author:   J. Zbiciak
 *  $Id: pads.c,v 1.16 1999/09/14 20:24:58 im14u2c Exp $
 * ============================================================================
 *  This module implements the controller pads.
 *  Pads are peripherals that extend periph_t.
 * ============================================================================
 */

static const char rcs_id[]="$Id: pads.c,v 1.16 1999/09/14 20:24:58 im14u2c Exp $";

#include "config.h"
#include "periph/periph.h"
#include "pads/pads.h"
#include "event/event.h"

#ifdef DIRECT_HCIF

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/io.h>

int pads_hcif_ok = 0;

/*
 * ===========================================================================
 *  INIT_PORTS   -- Get access to the desired I/O ports and drop root.
 *                  Don't bother returning an error if we fail.  Just abort.
 * ===========================================================================
 */
static int init_ports(uint_32 base)
{
    int bit = 0;

    /* -------------------------------------------------------------------- */
    /*  Convert the I/O base into a bitfield bit number.                    */
    /* -------------------------------------------------------------------- */
    switch (base)
    {
        case 0x378: bit = 1; break;
        case 0x278: bit = 2; break;
        case 0x3BC: bit = 4; break;
        default:    bit = 0; break;
    }

    /* -------------------------------------------------------------------- */
    /*  Complain if this seems to be an invalid port number.                */
    /* -------------------------------------------------------------------- */
    if (!bit)
    {
        fprintf(stderr, "direct_hcif:  Invalid base address 0x%.4X\n", base);
        return -1;
    }

    /* -------------------------------------------------------------------- */
    /*  See if we were able to get permissions on the requested I/O port.   */
    /* -------------------------------------------------------------------- */
    if ((bit & pads_hcif_ok) == 0)
    {
        fprintf(stderr, "direct_hcif:  Unable to set I/O permissions\n");
        return -1;
    }

    return 0;
}

/*
 * ============================================================================
 *  PAD_TK_HCIF  -- Reads from a dummy pad
 * ============================================================================
 */
sint_32 pad_tk_hcif(periph_t *p, sint_32 len)
{
    pad_t   *pad  = (pad_t*)p;
    uint_32 base  = pad->io_base;
    int     state = pad->rd_state++ & 7;
    int     side  = (state & 4) == 4;
    uint_32 val   = pad->rd_val;

    /* -------------------------------------------------------------------- */
    /*  Synchronous delays while we read the HCIF are bad.  Instead, we     */
    /*  have a four-phase state machine that moves between states at a low  */
    /*  enough rate while letting the rest of the emulator run in the       */
    /*  meantime.                                                           */
    /*                                                                      */
    /*  The state machine has four phases, and alternates between left and  */
    /*  right hand-controllers for a total of eight unique phases.          */
    /*                                                                      */
    /*  This code is almost a total rip-off from test_hcif.c.               */
    /* -------------------------------------------------------------------- */
    
    switch (state & 3)
    {
        /* ---------------------------------------------------------------- */
        /*  Get low nybble.                                                 */
        /* ---------------------------------------------------------------- */
        case 0:
            /* ------------------------------------------------------------ */
            /*  Phase 0:  Select low-nybble for this controller side.       */
            /* ------------------------------------------------------------ */
            outb((2<<side)|1, base);        
            break;
        
        case 1:
            /* ------------------------------------------------------------ */
            /*  Phase 1:  Read it back and mask it.                         */
            /* ------------------------------------------------------------ */
            val  = (inb(base + 1) >> 4) & 0x0F;
            break;

        /* ---------------------------------------------------------------- */
        /*  Get high nybble.                                                */
        /* ---------------------------------------------------------------- */
        case 2:
            /* ------------------------------------------------------------ */
            /*  Phase 2:  Select high-nybble for this controller side.      */
            /* ------------------------------------------------------------ */
            outb((2<<side)|0, base); 
            break;

        case 3:
            /* ------------------------------------------------------------ */
            /*  Phase 3:  Read it back, mask & merge it.  Then update the   */
            /*            controller's value in our pad_t struct.           */
            /* ------------------------------------------------------------ */
            val |= (inb(base + 1)     ) & 0xF0;

            pad->side[side] = val ^ 0x88;
            break;

        /* ---------------------------------------------------------------- */
        /*  Uhoh, something bad happened.                                   */
        /* ---------------------------------------------------------------- */
        default:
            fprintf(stderr, "Can't get here in pad_hcif_tick()\n");
            exit(1);
    }

    pad->rd_val = val;

    return len;
}

#endif /* DIRECT_HCIF */

/*
 * ============================================================================
 *  PAD_TK_EVENT -- Updates that pad's states according to event inputs.
 * ============================================================================
 */
sint_32 pad_tk_event(periph_t *p, sint_32 len)
{
    pad_t   *pad = (pad_t*)p;
    int     i, s;
    uint_32 value = 0, dflag;
    uint_32 *ptr;

    /* -------------------------------------------------------------------- */
    /*  If no input events have occurred, don't reevaluate the pads.        */
    /* -------------------------------------------------------------------- */
    if (pad->last_evt == event_count)
        return len;

    pad->last_evt = event_count;

    /* -------------------------------------------------------------------- */
    /*  Iterate over both controllers.                                      */
    /* -------------------------------------------------------------------- */
    for (s = 0; s < 2; s++)
    {
        /* ---------------------------------------------------------------- */
        /*  Start out with "nothing pressed."                               */
        /* ---------------------------------------------------------------- */
        value = 0;

        /* ---------------------------------------------------------------- */
        /*  Merge all of the keypad / action key inputs to the controller.  */
        /* ---------------------------------------------------------------- */
        ptr = (s) ? pad->l : pad->r;

        for (i = 0; i < 15; i++)
            value |= ptr[i];

        /* ---------------------------------------------------------------- */
        /*  Now, generate a disc dir # from E/NE/N/NW/W/SW/S/SE flags.      */
        /*                                                                  */
        /*  Input bits 0, 2, 4, and 6 give us pure E, N, W, and S.          */
        /*  Input bits 1, 3, 5, and 7 give us NE, NW, SE, SW.               */
        /*                                                                  */
        /*  Pad bit 0 is set for WSW through SE.                            */
        /*  Pad bit 1 is set for SSE through NE.                            */
        /*  Pad bit 2 is set for ENE through NW.                            */
        /*  Pad bit 3 is set for NNW through SW.                            */
        /*  Pad bit 4 is set for NE, NNE, NW, WNW, SW, SSW, SE, ESE.        */
        /*                                                                  */
        /*  Input bit to compass headings:                                  */
        /*                                                                  */
        /*      Compass         Pattern 1           Pattern 2               */
        /*         E            00000001                                    */
        /*         ENE          00000011                                    */
        /*         NE           00000010            00000101                */
        /*         NNE          00000110                                    */
        /*         N            00000100                                    */
        /*         NNW          00001100                                    */
        /*         NW           00001000            00010100                */
        /*         WNW          00011000                                    */
        /*         W            00010000                                    */
        /*         WSW          00110000                                    */
        /*         SW           00100000            01010000                */
        /*         SSW          01100000                                    */
        /*         S            01000000                                    */
        /*         SSE          11000000                                    */
        /*         SE           10000000            01000001                */
        /*         ESE          10000001                                    */
        /*                                                                  */
        /* ---------------------------------------------------------------- */
        dflag = (0xFF & ptr[15]);
        dflag = (dflag << 4) | (dflag >> 4);

        /* ---------------------------------------------------------------- */
        /*  Step through the four major compass dirs and set bits 0..3      */
        /*  according to each range.  Also process bit 4 along the way.     */
        /*  We begin our analysis with 'south' as that is bit 0.            */
        /* ---------------------------------------------------------------- */
        for (i = 0; i < 4; i++, dflag >>= 2)
        {
            /* ------------------------------------------------------------ */
            /*  Handle major-direction bit.  We set the bit for any dir     */
            /*  that is mostly in the direction of the major-direction, as  */
            /*  well as for one lop-sided case on the side.                 */
            /* ------------------------------------------------------------ */
            if ((dflag&0x14) != 0x10 && (dflag&0x0F) > 0x01) value |= 1 << i;

            /* ------------------------------------------------------------ */
            /*  Check for diagonal bit (bit 4) also.                        */
            /* ------------------------------------------------------------ */
            if ((0x64 >> (dflag & 0x07)) & 1)                value |= 0x10;
        }

        /* ---------------------------------------------------------------- */
        /*  Remember our output value.                                      */
        /* ---------------------------------------------------------------- */
        pad->side[s]     = 0xFF & ~value;
    }

    return len;
}

/*
 * ============================================================================
 *  PAD_READ     -- Returns the current state of the pads.
 * ============================================================================
 */
uint_32 pad_read        (periph_t *p, periph_t *r, uint_32 a, uint_32 d)
{
    pad_t *pad = (pad_t*)p;

    (void)r;    /* Unused */
    (void)d;    /* Unused */

    /* -------------------------------------------------------------------- */
    /*  Ignore accesses that are outside our address space.                 */
    /* -------------------------------------------------------------------- */
    if (a < 14) return ~0U;

    return (pad->side[a & 1] & 0xFF);
}

/*
 * ============================================================================
 *  PAD_INIT     -- Makes a dummy pad
 * ============================================================================
 */
int pad_init
(
    pad_t           *pad,       /*  mem_t structure to initialize   */
    uint_32         addr,       /*  Base address of pad.            */
    uint_32         io_base     /*  I/O address of HCIF, or 0.      */
)
{
    if (io_base > 0)
    {
#ifdef DIRECT_HCIF
        /* ---------------------------------------------------------------- */
        /*  Note:  This will only really work reliably if the emulator is   */
        /*  running around 100% or slower on an otherwise unloaded machine. */
        /*  I really need to check the _actual_ elapsed time in the _tick   */
        /*  function since sudden bursts of reads might cause _tick to be   */
        /*  called too quickly in succession.                               */
        /* ---------------------------------------------------------------- */
        if (init_ports(io_base))
            return -1;

        pad->periph.read     = pad_read;
        pad->periph.write    = NULL;
        pad->periph.tick     = pad_tk_hcif;
        pad->periph.min_tick = 3579545 / 960;   /* 60Hz scanning rate.  */
        pad->periph.max_tick = 3579545 / 960;   /* (times 4 states)     */

        pad->io_base         = io_base;       
        pad->side[0]         = 0xFF;
        pad->side[1]         = 0xFF;
        pad->rd_state        = 0;
        pad->rd_val          = 0;
#else
        fprintf(stderr, 
                "Error:  Direct Hand-Controller Interface support "
                "is not compiled in.\n");
        return -1;
#endif
    } else
    {
        pad->periph.read     = pad_read;
        pad->periph.write    = NULL;
        pad->periph.tick     = pad_tk_event;
        pad->periph.min_tick = 3579545 / 240;   /* 60Hz scanning rate.  */
        pad->periph.max_tick = 3579545 / 160;   /* 40Hz scanning rate.  */

        pad->rd_state        = 0;

        pad->side[0]         = 0xFF;
        pad->side[1]         = 0xFF;
    }

    pad->periph.addr_base = addr;
    pad->periph.addr_mask = 0xF;

    return 0;
}


/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */

